// zoomasm -- zoom video assembler
// (c) 2019,2020,2021,2022 Claude Heiland-Allen
// SPDX-License-Identifier: AGPL-3.0-or-later

#include "timeline.h"

#include "imgui.h"
#include "imfilebrowser.h"

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <cmath>
#include <exception>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>

#include "resource.h"
#include "utility.h"

class invalid_data : public std::exception
{
  virtual const char *what() const noexcept
  {
    return "invalid data";
  }
};

extern void seconds_to_timespec(double time, int timespec[4])
{
  int hours = std::floor(time / 60.0 / 60.0);
  int minutes = std::floor((time - 60.0 * 60.0 * hours) / 60.0);
  int seconds = std::floor(time - 60.0 * 60.0 * hours - 60.0 * minutes);
  int milliseconds = std::round(1000.0 * (time - 60.0 * 60.0 * hours - 60.0 * minutes - seconds));
  timespec[0] = hours;
  timespec[1] = minutes;
  timespec[2] = seconds;
  timespec[3] = milliseconds;
}

extern double timespec_to_seconds(const int timespec[4])
{
  int hours = timespec[0];
  int minutes = timespec[1];
  int seconds = timespec[2];
  int milliseconds = timespec[3];
  return ((((hours * 60.0) + minutes) * 60.0) + seconds) + milliseconds / 1000.0;
}

extern std::string format_timespec(const int timespec[4])
{
  std::ostringstream s;
  if (timespec[0])
  {
    s <<                                      timespec[0] << ':'
      << std::setw(2) << std::setfill('0') << timespec[1] << ':'
      << std::setw(2) << std::setfill('0') << timespec[2] << '.'
      << std::setw(3) << std::setfill('0') << timespec[3];
  }
  else if (timespec[1])
  {
    s <<                                      timespec[1] << ':'
      << std::setw(2) << std::setfill('0') << timespec[2] << '.'
      << std::setw(3) << std::setfill('0') << timespec[3];
  }
  else
  {
    s <<                                      timespec[2] << '.'
      << std::setw(3) << std::setfill('0') << timespec[3];
  }
  return s.str();
}

extern std::string format_time(double time)
{
  int timespec[4] = { 0, 0, 0, 0 };
  seconds_to_timespec(time, timespec);
  return format_timespec(timespec);
}

extern std::string format_time(const struct timeline *tl, double fraction)
{
  return format_time(fraction * timeline_get_duration(tl));
}

const char *interpolation_mode_string[3] = { "Step", "Linear", "Cubic" };

struct timeline
{
  double duration;
  double time;
  timestep cursor;
  std::vector<timestep> steps;
  ImGui::FileBrowser *export_csv;
  ImGui::FileBrowser *import_csv;
  bool export_interpolated;
  bool headerless;
};

extern struct timeline *timeline_new(void)
{
  struct timeline *tl = new timeline();
  if (! tl)
  {
    return nullptr;
  }
  tl->duration = 1;
  tl->time = 0;
  tl->cursor = timestep{0, 0, interpolation_cubic};
  tl->steps.push_back(timestep{0, 0, interpolation_cubic});
  tl->steps.push_back(timestep{1, 0, interpolation_cubic});
  tl->export_csv = new ImGui::FileBrowser(ImGuiFileBrowserFlags_CloseOnEsc | ImGuiFileBrowserFlags_EnterNewFilename | ImGuiFileBrowserFlags_CreateNewDir);
  tl->export_csv->SetTitle("Export timeline");
  tl->export_csv->SetTypeFilters({ ".csv" });
  tl->import_csv = new ImGui::FileBrowser(ImGuiFileBrowserFlags_CloseOnEsc);
  tl->import_csv->SetTitle("Import timeline");
  tl->import_csv->SetTypeFilters({ ".csv" });
  return tl;
}

extern void timeline_delete(struct timeline *tl)
{
  if (tl)
  {
    if (tl->export_csv)
    {
      delete tl->export_csv;
      tl->export_csv = nullptr;
    }
    if (tl->import_csv)
    {
      delete tl->import_csv;
      tl->import_csv = nullptr;
    }
    delete tl;
  }
}

extern double timeline_get_duration(const struct timeline *tl)
{
  return tl->duration;
}

extern bool timeline_set_duration(struct timeline *tl, double duration)
{
  if (duration > 0)
  {
    tl->duration = duration;
    tl->time = std::min(tl->time, tl->duration);
    return true;
  }
  return false;
}

extern double timeline_get_time(const struct timeline *tl)
{
  return tl->time;
}

extern bool timeline_set_time(struct timeline *tl, double time)
{
  tl->time = std::min(std::max(0.0, time), tl->duration);
  return true;
}

static void timeline_remove_timestep(struct timeline *tl, double fraction)
{
  // need to update phase of endpoints, which is done via remove+insert
  // so don't test for bounds
  // if (0.0 < fraction && fraction < 1.0)
  {
    for (int i = 0; i < (int) tl->steps.size(); ++i)
    {
      if (tl->steps[i].fraction == fraction)
      {
        // delete all items with matching fraction
        tl->steps.erase(tl->steps.begin() + i--);
      }
    }
  }
}

extern bool timeline_insert_timestep(struct timeline *tl, const struct timestep &ts)
{
  // need to update phase of endpoints, which is done via remove+insert
  // so don't test for bounds
  // if (0.0 < ts.fraction && ts.fraction < 1.0)
  {
    for (auto it = tl->steps.begin(); it != tl->steps.end(); ++it)
    {
      if (ts.fraction < it->fraction)
      {
        tl->steps.insert(it, ts);
        return true;
      }
    }
    tl->steps.insert(tl->steps.end(), ts);
    return true;
  }
  // assert(! "timeline_insert_timestep");
}

static void timestep_display(struct timeline *tl, int id, struct timestep ts, int input_keyframe_count, bool is_cursor, bool is_endpoint, bool enter_key_pressed, int NUMBER_OF_LAYERS, struct audio *au)
{
  ImGui::PushID(id);
  if (is_cursor)
  {
    ImGui::PushStyleColor(ImGuiCol_FrameBg, button_red);
  }
  double time = timeline_get_duration(tl) * ts.fraction;
  ImGui::PushItemWidth(100);
  if (is_cursor)
  {
    ImGui::Text("%9.3fs", time);
  }
  else
  {
    char label[100];
    snprintf(label, 100, "%9.3fs", time);
    if (ImGui::Button(label))
    {
      audio_set_time(au, time);
    }
  }
  ImGui::PopItemWidth();
  ImGui::SameLine();
  ImGui::PushItemWidth(250);
  float phasef = ts.phase;
  if (ImGui::SliderFloat("Z", &phasef, -(NUMBER_OF_LAYERS - 1), input_keyframe_count - 1 + (NUMBER_OF_LAYERS - 1), "%.3f"))
  {
    if (! is_cursor)
    {
      timeline_remove_timestep(tl, ts.fraction);
      ts.phase = phasef;
      timeline_insert_timestep(tl, ts);
    }
  }
  if (ImGui::IsItemHovered())
  {
    if (is_cursor)
    {
      ImGui::SetTooltip("Show zoom depth (keyframe index) at current time.");
    }
    else
    {
      ImGui::SetTooltip("Set zoom depth (keyframe index) for this waypoint.");
    }
  }
  ImGui::PopItemWidth();
  ImGui::SameLine();
  ImGui::PushItemWidth(100);
  int interp = ts.mode;
  if (ImGui::Combo("##Interpolation", &interp, interpolation_mode_string, IM_ARRAYSIZE(interpolation_mode_string)))
  {
    if (! is_cursor)
    {
      timeline_remove_timestep(tl, ts.fraction);
      ts.mode = (enum interpolation_mode) interp;
      timeline_insert_timestep(tl, ts);
    }
  }
  if (ImGui::IsItemHovered())
  {
    if (is_cursor)
    {
      ImGui::SetTooltip("Show interpolation type starting at this waypoint.");
    }
    else
    {
      ImGui::SetTooltip("Set interpolation type starting at this waypoint.");
    }
  }
  ImGui::PopItemWidth();
  if (is_cursor)
  {
    ImGui::SameLine();
    if (ImGui::Button("+") || enter_key_pressed)
    {
      timeline_insert_timestep(tl, ts);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Add a new waypoint here.");
    }
  }
  else if (! is_endpoint)
  {
    ImGui::SameLine();
    if (ImGui::Button("-"))
    {
      timeline_remove_timestep(tl, ts.fraction);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Delete this waypoint permanently.  There is no undo.");
    }
  }
  if (is_cursor)
  {
    ImGui::PopStyleColor(1);
  }
  ImGui::PopID();
}

extern void timeline_display_early(struct timeline *tl, int input_keyframe_count, bool enter_key_pressed, bool session_loaded, int NUMBER_OF_LAYERS, struct audio *au)
{
  ImGui::Begin("Timeline", nullptr, ImGuiWindowFlags_HorizontalScrollbar);
  if (ImGui::Button("Import"))
  {
    tl->import_csv->Open();
  }
  if (ImGui::IsItemHovered())
  {
    ImGui::SetTooltip("Import timeline from CSV text file.");
  }
  ImGui::SameLine();
  if (ImGui::Button("Export"))
  {
    tl->export_csv->Open();
  }
  if (ImGui::IsItemHovered())
  {
    ImGui::SetTooltip("Export timeline to CSV text file.");
  }
  ImGui::SameLine();
  ImGui::Checkbox("Headerless", &tl->headerless);
  if (ImGui::IsItemHovered())
  {
    ImGui::SetTooltip("Set this if CSV files do/should not have a header line.");
  }
  ImGui::SameLine();
  ImGui::Checkbox("Interpolated", &tl->export_interpolated);
  if (ImGui::IsItemHovered())
  {
    ImGui::SetTooltip("Export interpolated timeline (uses Output/FPS, cannot be imported).");
  }
  // update last waypoint if the input count changed
  static int input_keyframe_count_old = -1;
  bool input_keyframe_count_changed = false;
  if (input_keyframe_count_old != input_keyframe_count)
  {
    // only auto-update if we didn't load a session
    if (! session_loaded)
    {
      input_keyframe_count_changed = true;
    }
    input_keyframe_count_old = input_keyframe_count;
  }
  // keyframe time steps
  double fraction = timeline_get_time(tl) / timeline_get_duration(tl);
  fraction = std::min(std::max(fraction, 0.0), 1.0);
  if (fraction != tl->cursor.fraction)
  {
    tl->cursor = timeline_interpolate(tl, fraction);
  }
  bool shown_cursor = false;
  for (int i = 0; i < (int) tl->steps.size(); ++i)
  {
    bool is_endpoint = i == 0 || i == (int) tl->steps.size() - 1;
    bool show_speed = false;
    if (fraction == tl->steps[i].fraction)
    {
      // don't show cursor if it's exactly on a keyframe
      shown_cursor = true;
      show_speed = true;
    }
    if (fraction < tl->steps[i].fraction && ! shown_cursor)
    {
      timestep_display(tl, -i, tl->cursor, input_keyframe_count, true, is_endpoint, enter_key_pressed, NUMBER_OF_LAYERS, au);
      shown_cursor = true;
      show_speed = true;
    }
    if (show_speed)
    {
      double dt = 0.01;
      double t0 = timeline_get_time(tl);
      double t1 = t0 + dt;
      double tt = timeline_get_duration(tl);
      double f0 = t0 / tt;
      double f1 = t1 / tt;
      double p0 = timeline_interpolate(tl, f0).phase;
      double p1 = timeline_interpolate(tl, f1).phase;
      double speed = (p1 - p0) / dt;
      ImGui::Text("Speed: %.3f 2x zooms per second", speed);
      ImGui::Text("Speed: %.3f 10x zooms per second", speed * std::log10(2.0));
    }
    // update last waypoint if the input count changed
    if (i == (int) tl->steps.size() - 1 && input_keyframe_count_changed)
    {
      struct timestep ts = tl->steps[i];
      timeline_remove_timestep(tl, ts.fraction);
      ts.phase = input_keyframe_count;
      timeline_insert_timestep(tl, ts);
    }
    timestep_display(tl, i, tl->steps[i], input_keyframe_count, false, is_endpoint, enter_key_pressed, NUMBER_OF_LAYERS, au);
  }
  ImGui::End();
}

extern void timeline_display_late(struct timeline *tl, double fps)
{
  static std::string error_string = "";
  tl->export_csv->Display();
  if (tl->export_csv->HasSelected())
  {
    std::filesystem::path file = tl->export_csv->GetSelected();
    tl->export_csv->ClearSelected();
    try
    {
      std::ofstream out(file); // FIXME handle exceptions
      out.imbue(std::locale::classic());
      if (tl->export_interpolated)
      {
        int output_frame_count = fps * tl->duration;
        if (! tl->headerless)
        {
          out << "Frame" << "," << "Z" << "\r\n";
        }
        for (int frame = 0; frame < output_frame_count; ++frame)
        {
          double fraction = frame / (double) output_frame_count;
          out << frame << "," << std::fixed << std::setprecision(6) << timeline_interpolate(tl, fraction).phase << "\r\n";
        }
      }
      else
      {
        if (! tl->headerless)
        {
          out << "Fraction" << "," << "Z" << "," << "Interpolation" << "\r\n";
        }
        for (int t = 0; t < (int) tl->steps.size(); ++t)
        {
          struct timestep ts = tl->steps[t];
          out << std::fixed << std::setprecision(6) << ts.fraction << ","
              << std::fixed << std::setprecision(6) << ts.phase << ","
              << interpolation_mode_string[ts.mode] << "\r\n";
        }
      }
    }
    catch (std::exception &e)
    {
      error_string = e.what();
    }
  }
  tl->import_csv->Display();
  if (tl->import_csv->HasSelected())
  {
    std::filesystem::path file = tl->import_csv->GetSelected();
    tl->import_csv->ClearSelected();
    std::vector<struct timestep> steps_old = tl->steps;
    tl->steps.clear();
    try
    {
      std::ifstream in(file);
      in.imbue(std::locale::classic());
      std::string line;
      if (! tl->headerless)
      {
        getline(in, line);
        if (line != "Fraction,Z,Interpolation")
        {
          throw invalid_data();
        }
      }
      while (! getline(in, line).eof())
      {
        if (line == "")
        {
          continue;
        }
        size_t pos = 0;
        std::string line0 = line;
        if ((pos = line.find(',')) != std::string::npos)
        {
          double fraction = std::stof(line.substr(0, pos));
          line.erase(0, pos + 1);
          if ((pos = line.find(',')) != std::string::npos)
          {
            double phase = std::stof(line.substr(0, pos));
            line.erase(0, pos + 1);
            enum interpolation_mode mode = interpolation_step;
            if (line == "Step" || line == "Linear" || line == "Cubic")
            {
              if (line == "Step")
              {
                mode = interpolation_step;
              }
              else if (line == "Linear")
              {
                mode = interpolation_linear;
              }
              else if (line == "Cubic")
              {
                mode = interpolation_cubic;
              }
              timeline_insert_timestep(tl, { fraction, phase, mode });
              continue;
            }
          }
        }
        throw invalid_data();
      }
      if (tl->steps.size() < 2 || tl->steps[0].fraction != 0 || tl->steps[tl->steps.size() - 1].fraction != 1)
      {
        throw invalid_data();
      }
    }
    catch (std::exception &e)
    {
      error_string = e.what();
      tl->steps = steps_old;
    }
  }
  if (error_string != "")
  {
    ImGui::OpenPopup("Error");
    ImVec2 center(ImGui::GetIO().DisplaySize.x * 0.5f, ImGui::GetIO().DisplaySize.y * 0.5f);
    ImGui::SetNextWindowPos(center, ImGuiCond_Appearing, ImVec2(0.5f, 0.5f));
    if (ImGui::BeginPopupModal("Error", NULL, ImGuiWindowFlags_AlwaysAutoResize))
    {
      ImGui::Text("%s", error_string.c_str());
      if (ImGui::Button("Ok"))
      {
        error_string = "";
      }
      ImGui::EndPopup();
    }
  }
}

extern timestep timeline_interpolate(struct timeline *tl, double fraction)
{
  // TODO monotone cubic interpolation
  fraction = std::min(std::max(fraction, 0.0), 1.0);
  int lo = 0;
  int hi = (int) tl->steps.size() - 1;
  if (! (tl->steps[lo].fraction < fraction))
  {
    return timestep{0, tl->steps[lo].phase, tl->steps[lo].mode};
  }
  if (! (fraction < tl->steps[hi].fraction))
  {
    return timestep{1, tl->steps[hi].phase, tl->steps[hi].mode};
  }
  assert(tl->steps[lo].fraction <= fraction);
  assert(fraction <= tl->steps[hi].fraction);
  // binary search for neighbours
  while (lo + 1 < hi)
  {
    int md = (lo + hi) / 2;
    assert(lo < md);
    assert(md < hi);
    assert(tl->steps[lo].fraction <= tl->steps[md].fraction);
    assert(tl->steps[md].fraction <= tl->steps[hi].fraction);
    if (fraction <= tl->steps[md].fraction)
    {
      hi = md;
    }
    else if (tl->steps[md].fraction <= fraction)
    {
      lo = md;
    }
    else
    {
      assert(! "nan fraction?");
    }
    assert(tl->steps[lo].fraction <= fraction);
    assert(fraction <= tl->steps[hi].fraction);
  }
  assert(lo + 1 == hi);
  enum interpolation_mode mode = tl->steps[lo].mode;
  switch (mode)
  {
    case interpolation_step:
    {
      double plo = tl->steps[lo].phase;
      double phase = plo;
      return timestep{ fraction, phase, mode };
    }
    case interpolation_linear:
    {
      double flo = tl->steps[lo].fraction;
      double plo = tl->steps[lo].phase;
      double fhi = tl->steps[hi].fraction;
      double phi = tl->steps[hi].phase;
      double t = (fraction - flo) / (fhi - flo);
      t = std::min(std::max(t, 0.0), 1.0);
      double phase = plo + t * (phi - plo);
      return timestep{ fraction, phase, mode };
    }
    case interpolation_cubic:
    {
      double tangent_lo = 0;
      if (0 < lo)
      {
        switch (tl->steps[lo - 1].mode)
        {
          case interpolation_step:
          {
            tangent_lo = 0;
            break;
          }
          case interpolation_linear:
          {
            tangent_lo
              = (tl->steps[lo].phase - tl->steps[lo - 1].phase)
              / (tl->steps[lo].fraction - tl->steps[lo - 1].fraction);
            break;
          }
          case interpolation_cubic:
          {
            double d_lo
              = (tl->steps[lo].phase - tl->steps[lo - 1].phase)
              / (tl->steps[lo].fraction - tl->steps[lo - 1].fraction);
            double d_hi
              = (tl->steps[hi].phase - tl->steps[lo].phase)
              / (tl->steps[hi].fraction - tl->steps[lo].fraction);
            if (d_lo == 0 || d_hi == 0 || (d_lo > 0) != (d_hi > 0))
            {
              tangent_lo = 0;
            }
            else
            {
              double m = (d_lo + d_hi) / 2;
              double a = m / d_lo;
              if (a < 0)
              {
                m = 0;
                a = 0;
              }
              if (a > 3)
              {
                m = 3 * d_lo;
              }
              tangent_lo = m;
            }
            break;
          }
        }
      }
      else
      {
        // secant
        tangent_lo
          = (tl->steps[hi].phase - tl->steps[lo].phase)
          / (tl->steps[hi].fraction - tl->steps[lo].fraction);
      }
      double tangent_hi = 0;
      if (hi < (int) tl->steps.size())
      {
        switch (tl->steps[hi].mode)
        {
          case interpolation_step:
          {
            tangent_hi = 0;
            break;
          }
          case interpolation_linear:
          {
            if (hi + 1 < (int) tl->steps.size())
            {
              tangent_hi
                = (tl->steps[hi + 1].phase - tl->steps[hi].phase)
                / (tl->steps[hi + 1].fraction - tl->steps[hi].fraction);
            }
            else
            {
              tangent_hi
                = (tl->steps[hi].phase - tl->steps[lo].phase)
                / (tl->steps[hi].fraction - tl->steps[lo].fraction);
            }
            break;
          }
          case interpolation_cubic:
          {
            double d_lo
              = (tl->steps[hi].phase - tl->steps[lo].phase)
              / (tl->steps[hi].fraction - tl->steps[lo].fraction);
            double d_hi;
            if (hi + 1 < (int) tl->steps.size())
            {
              d_hi
                = (tl->steps[hi + 1].phase - tl->steps[hi].phase)
                / (tl->steps[hi + 1].fraction - tl->steps[hi].fraction);
            }
            else
            {
              d_hi = d_lo;
            }
            if (d_lo == 0 || d_hi == 0 || (d_lo > 0) != (d_hi > 0))
            {
              tangent_hi = 0;
            }
            else
            {
              double m = (d_lo + d_hi) / 2;
              double b = m / d_hi;
              if (b < 0)
              {
                m = 0;
                b = 0;
              }
              if (b > 3)
              {
                m = 3 * d_hi;
              }
              tangent_hi = m;
            }
            break;
          }
        }
      }
      // cubic Hermite spline
      double flo = tl->steps[lo].fraction;
      double plo = tl->steps[lo].phase;
      double fhi = tl->steps[hi].fraction;
      double phi = tl->steps[hi].phase;
      double t = (fraction - flo) / (fhi - flo);
      double h00 = (1 + 2 * t) * (1 - t) * (1 - t);
      double h10 = t * (1 - t) * (1 - t);
      double h01 = t * t * (3 - 2 * t);
      double h11 = t * t * (t - 1);
      t = std::min(std::max(t, 0.0), 1.0);
      tangent_lo *= fhi - flo;
      tangent_hi *= fhi - flo;
      double phase = h00 * plo + h10 * tangent_lo + h01 * phi + h11 * tangent_hi;
      return timestep{ fraction, phase, mode };
    }
  }
  assert(! "interpolation mode");
}

extern int timeline_get_count(const struct timeline *tl)
{
  return tl->steps.size();
}

extern struct timestep timeline_get_timestep(const struct timeline *tl, int ix)
{
  assert(0 <= ix && ix < (int) tl->steps.size());
  return tl->steps[ix];
}

extern bool timeline_clear(struct timeline *tl)
{
  tl->steps.clear();
  return true;
}

extern bool timeline_reset(struct timeline *tl)
{
  timeline_clear(tl);
  timeline_insert_timestep(tl, { 0.0, 0.0, interpolation_cubic });
  timeline_insert_timestep(tl, { 1.0, 0.0, interpolation_cubic });
  return true;
}

extern bool timeline_is_valid(const struct timeline *tl)
{
  if (tl && tl->steps.size() >= 2)
  {
    return tl->steps.front().fraction == 0.0 && tl->steps.back().fraction == 1.0;
  }
  return false;
}
